{****************************************************************************

                  Copyright (c) 1993,94 by Florian Klmpfl

 ****************************************************************************}

unit asmgen;

  interface

    uses
       globals,cobjects,strings;

    type
       tasmop = (MOV,MOVZX,MOVSX,A_GLOBAL,A_LABEL,A_STATIC,ASCII,ADD,
                 CALL,DIRECT,IDIV,IMUL,JMP,LEA,MUL,NEG,A_NOT,
                 POP,POPAD,PUSH,PUSHAD,RET,SUB,XCHG,A_XOR,
                 FILD,CMP,JZ,A_INC,A_DEC,SETE,SETNE,SETL,
                 SETG,SETLE,SETGE,JE,JNE,JL,JG,JLE,JGE,
                 A_OR,FLD,FADD,FMUL,FSUB,FDIV,FCHS,FLD1,
                 FIDIV,CLTD,JNZ,FSTP,A_DOUBLE,A_AND,JNO,A_NOTH,A_NONE,
                 ENTER,LEAVE,CLD,MOVS,REP,A_SHL,A_SHR,A_LONG,A_BOUND,
                 A_BYTE,A_WORD,JNS,JS,JO,SAR,TEST,STABS,STABD,
                 FCOM,FCOMP,FCOMPP,FXCH,FADDP,FMULP,FSUBP,FDIVP,
                 FNSTS,SAHF);

       topsize = (S_NO,S_B,S_W,S_L,S_BW,S_BL,S_WL);

       { Aufzhlungstyp fr die Register }
       tregister = (R_NO,R_EAX,R_EBX,R_ECX,R_EDX,R_ESI,R_EDI,R_EBP,R_ESP,
                    R_AX,R_BX,R_CX,R_DX,R_SI,R_DI,R_AL,R_BL,R_CL,R_DL);

       tsetofregisters =set of tregister;

       tdisposing = (D_LABEL,D_STRING);

       preferenz = ^treferenz;

       treferenz = record
          isintvalue : boolean;
          base,index : tregister;
          scalefactor : byte;
          offset : longint;
          symbol : pstring;
       end;

       tsimplasmoptyp = (IS_MEM,IS_REG,IS_CONST);

       tsimplasmop = record
          case optyp : tsimplasmoptyp of
             IS_MEM : (ref : treferenz);
             IS_REG : (reg : tregister);
       end;

       pasmrec = ^tasmrec;

       tasmrec = record
          next : pasmrec;
          opsize : topsize;
          instruc : tasmop;
          linenr : longint;
          infile : pinputstack;
          case disposing : tdisposing of
             D_LABEL : (l : longint);
             D_STRING : (t : pchar);
       end;

       pasmlist = ^tasmlist;

       tasmlist = object
          wurzel,last : pasmrec;
          constructor init;
          destructor done;

          { wird am Ende angehngt }
          procedure concat(p : pasmrec);
          procedure concatlist(p : pasmlist);

          { fgen am Anfang ein }
          procedure insert(p : pasmrec);
          procedure insertlist(p : pasmlist);
       end;

    const
       regid2str : array[R_NO..R_DL] of string[4] =
          ('','%eax','%ebx','%ecx','%edx','%esi','%edi','%ebp','%esp',
           '%ax','%bx','%cx','%dx','%si','%di','%al','%bl','%cl','%dl');

    procedure asmgeninit;
    function gennasmrec(i : tasmop;s : topsize;const t : string) : pasmrec;
    function genlasmrec(i : tasmop;l : longint) : pasmrec;
    function genpasmrec(i : tasmop;s : topsize;pc : pchar) : pasmrec;
    procedure writemainasmlist;
    procedure disposeasmrec(p : pasmrec);
    function getreferenzstring(const ref : treferenz) : string;
    procedure clear_referenz(var ref : treferenz);
    function tolabel(l : longint) : string;

    { fr den Sprungoptimierer }
    type
       pusedinfo = ^tusedinfo;

       tusedinfo = record
          used_at : pasmrec;
          next : pusedinfo;
       end;

       plabelinfo = ^tlabelinfo;

       tlabelinfo = record
          defined : pasmrec;
          used : pusedinfo;
          next : plabelinfo;
       end;

    procedure registerlabel(l : longint);
    function getinforec(l : longint) : plabelinfo;
    procedure delused(p : pasmrec);
    procedure registerused(p : pasmrec;l : longint);

    { gibt eine eindeutige Nummer fr ein Label zurck }
    function getlabel : longint;

    { eindeutiges Label fuer einen Konstante }
    function getconstlabel : longint;

    { ... hier aber nicht registriert }
    function getunreglabel : longint;

    { Registerkonvertierungen }
    function reg8toreg16(reg : tregister) : tregister;
    function reg8toreg32(reg : tregister) : tregister;
    function reg16toreg8(reg : tregister) : tregister;
    function reg32toreg8(reg : tregister) : tregister;
    function reg32toreg16(reg : tregister) : tregister;
    function reg16toreg32(reg : tregister) : tregister;

    var
       mainasmlist : tasmlist;

       { asmlist fuer das Datensegment }
       datasegment : tasmlist;

       { asmlist fuer die Konstanten }
       constsegment : tasmlist;

       { gibt die Anzahl der erzeugten Zeilen an }
       asmlines : longint;

       { ist true, wenn ein push oder pusha generiert wurde }
       pushgened : boolean;

       { Assemblerliste, in die die VMT's eingefgt werden }
       vmtasmlist : tasmlist;

  implementation

    var
       { Zhler zum Erzeugen von Labels }
       labelcount : longint;

    function reg8toreg16(reg : tregister) : tregister;

      begin
         reg8toreg16:=reg32toreg16(reg8toreg32(reg));
      end;

    function reg16toreg8(reg : tregister) : tregister;

      begin
         reg16toreg8:=reg32toreg8(reg16toreg32(reg));
      end;

    function reg16toreg32(reg : tregister) : tregister;

      begin
         reg16toreg32:=tregister(byte(reg)-byte(R_ESP));
      end;

    function reg32toreg16(reg : tregister) : tregister;

      begin
         reg32toreg16:=tregister(byte(reg)+byte(R_ESP));
      end;

    function reg32toreg8(reg : tregister) : tregister;

      begin
         reg32toreg8:=tregister(byte(reg)+byte(R_DI));
      end;

    function reg8toreg32(reg : tregister) : tregister;

      begin
         reg8toreg32:=tregister(byte(reg)-byte(R_DI));
      end;

    function getlabel : longint;

      begin
         inc(labelcount);
{         if jumpoptimize then
           registerlabel(labelcount); }
         getlabel:=labelcount;
      end;

    function getconstlabel : longint;

      begin
         inc(labelcount);
         getconstlabel:=labelcount;
      end;

    function getunreglabel : longint;

      begin
         inc(labelcount);
         getunreglabel:=labelcount;
      end;
      
    { Verwaltung der Labelpositionen (fr Sprungoptimierungen) }

    const
       labelcontsize = 500;

    type
       tlabelcontainer = array[0..labelcontsize-1] of plabelinfo;

    var
       labelcontainer : ^tlabelcontainer;

    function getinforec(l : longint) : plabelinfo;

      var
         hp : plabelinfo;
         i : word;

      begin
         hp:=labelcontainer^[l mod labelcontsize];
         for i:=1 to (l div labelcontsize) do
           hp:=hp^.next;
         getinforec:=hp;
      end;

    procedure registerlabel(l : longint);

      var
         hp : plabelinfo;

      begin
         new(hp);
         hp^.next:=labelcontainer^[l mod labelcontsize];
         hp^.used:=nil;
         hp^.defined:=nil;
         labelcontainer^[l mod labelcontsize]:=hp;
      end;

    procedure delused(p : pasmrec);

      var
         hp1 : plabelinfo;
         hp2,hp3 : pusedinfo;
         l : longint;

      begin
         l:=p^.l;
         hp1:=getinforec(l);
         hp2:=hp1^.used;
         if hp2^.used_at=p then
           begin
              hp3:=hp2^.next;
              dispose(hp2);
              hp1^.used:=hp3;
           end
         else
           begin
              while hp2^.next^.used_at<>p do
                hp2:=hp2^.next;
              hp3:=hp2^.next^.next;
              dispose(hp2^.next);
              hp2^.next:=hp3;
           end;
      end;

    procedure registerused(p : pasmrec;l : longint);

      var
         hp1 : plabelinfo;
         hp2 : pusedinfo;

      begin
         hp1:=getinforec(l);
         new(hp2);
         hp2^.used_at:=p;
         hp2^.next:=hp1^.used;
         hp1^.used:=hp2;
      end;

    function getreferenzstring(const ref : treferenz) : string;

      var
         s : string;

      begin
         if ref.isintvalue then
           s:='$'+tostr(ref.offset)
         else
{$ifdef tp}
           with ref do
             begin
                s:='';
                if assigned(symbol) then
                  s:=symbol^;
                if offset<0 then s:=s+tostr(offset)
                  else if (offset>0) then
                    begin
                       if (symbol=nil) then s:=tostr(offset)
                       else s:=s+'+'+tostr(offset);
                    end;
                if (index<>R_NO) and (base=R_NO) then
                  s:=s+'(,'+regid2str[index]+','+tostr(scalefactor)+')'
                else if (index=R_NO) and (base<>R_NO) then
                  s:=s+'('+regid2str[base]+')'
                else if (index<>R_NO) and (base<>R_NO) then
                  s:=s+'('+regid2str[base]+','
                    +regid2str[index]+','+tostr(scalefactor)+')';
             end;
{$else}
           begin
              s:='';
              if assigned(ref.symbol) then
                s:=ref.symbol^;
              if ref.offset<0 then s:=s+tostr(ref.offset)
                else if (ref.offset>0) then
                  begin
                     if (ref.symbol=nil) then s:=tostr(ref.offset)
                     else s:=s+'+'+tostr(ref.offset);
                  end;
              if (ref.index<>R_NO) and (ref.base=R_NO) then
                s:=s+'(,'+regid2str[ref.index]+','+tostr(ref.scalefactor)+')'
              else if (ref.index=R_NO) and (ref.base<>R_NO) then
                s:=s+'('+regid2str[ref.base]+')'
              else if (ref.index<>R_NO) and (ref.base<>R_NO) then
                s:=s+'('+regid2str[ref.base]+','
                  +regid2str[ref.index]+','+tostr(ref.scalefactor)+')';
           end;
{$endif}
         getreferenzstring:=s;
      end;

    procedure clear_referenz(var ref : treferenz);

      begin
{$ifdef tp}
         with ref do
           begin
              index:=R_NO;
              base:=R_NO;
              offset:=0;
              scalefactor:=1;
              symbol:=nil;
              isintvalue:=false;
           end;
{$else}
         ref.index:=R_NO;
         ref.base:=R_NO;
         ref.offset:=0;
         ref.scalefactor:=1;
         ref.symbol:=nil;
         ref.isintvalue:=false;
{$endif}
      end;

    { ******************************************************** }

    const
       op_2_asm : array[MOV..SAHF] of string[7] =
       ('mov','movz','movs','.comm','','.lcomm','.ascii','add',
        'call','','idiv','imul','jmp','lea','mul','neg','not',
        'pop','popal','push','pushal','ret','sub','xchg','xor',
        'fild','cmp','jz','inc','dec','sete','setne','setl',
        'setg','setle','setge','je','jne','jl','jg','jle','jge',
        'or','fld','fadd','fmul','fsub','fdiv','fchs','fld1',
        'fidiv','cltd','jnz','fstp','.double','and','jno','','',
        'enter','leave','cld','movs','rep','shl','shr','.long','bound',
        '.byte','.word','jns','js','jo','sar','test','.stabs','.stabd',
        'fcom','fcomp','fcompp','fxch','faddp','fmulp','fsubp','fdivp',
        'fnsts','sahf');

    const postf_2_asm : array[S_NO..S_WL] of string[2] = ('','b','w','l','bw',
                                                       'bl','wl');

    constructor tasmlist.init;

      begin
         wurzel:=nil;
         last:=nil;
      end;

    destructor tasmlist.done;

      var
         hp : pasmrec;

      begin
         hp:=wurzel;
         while assigned(hp) do
           begin
              wurzel:=hp^.next;
              disposeasmrec(hp);
              hp:=wurzel;
           end;
      end;

    procedure tasmlist.insert(p : pasmrec);

      begin
         p^.next:=wurzel;
         wurzel:=p;
         if last=nil then last:=p;
      end;

    procedure tasmlist.insertlist(p : pasmlist);

      begin
         p^.last^.next:=wurzel;
         wurzel:=p^.wurzel;
         if last=nil then last:=p^.last;
         p^.last:=nil;
         p^.wurzel:=nil;
      end;

    procedure tasmlist.concat(p : pasmrec);

      begin
         if wurzel=nil then wurzel:=p
           else last^.next:=p;
         last:=p;
      end;

    procedure tasmlist.concatlist(p : pasmlist);

     begin
         if p^.wurzel=nil then exit;
         if wurzel=nil then wurzel:=p^.wurzel
           else last^.next:=p^.wurzel;
         last:=p^.last;
         p^.last:=nil;
         p^.wurzel:=nil;
      end;

    procedure disposeasmrec(p : pasmrec);

      begin
         if p^.disposing=D_STRING then strdispose(p^.t);
         dispose(p);
      end;

    procedure asmgeninit;

      procedure writeasm(s : string);

        begin
           mainasmlist.concat(gennasmrec(DIRECT,S_NO,'//  '+s));
        end;

      function janein(b : boolean) : string;

        begin
           if b then janein:='Ja'
             else janein:='Nein';
        end;

      var
         i : integer;

      begin
         labelcount:=0;
         mainasmlist.init;
         writeasm('Uebersetzt von FPKPascal  '+version);
         writeasm('');
         writeasm('Zielbetriebssystem:');
         if target=target_DOS then
           writeasm('    DOS-Extender')
         else if target=target_LINUX then
           writeasm('    Linux')
         else if target=target_OS2 then
           writeasm('    OS/2 2.x')
         else if target=target_WIN32 then
           writeasm('    Windows 32 Bit');
         writeasm('');
         writeasm('Compileroptionen:');
         if littlesize then
           writeasm('    Grenoptimierung')
         else writeasm('    Geschwindigkeitsoptimierung');
         writeasm('    Einfache Optimierungen:      '+janein(optimiere));
         writeasm('    Ueberlaufpruefung:           '+janein(check_overflow));
         writeasm('    Bereichsueberpruefung:       '+janein(rangechecking));
         writeasm('    Ausnahmebehandlung:          '+janein(genexceptcode));
         mainasmlist.concat(gennasmrec(DIRECT,S_NO,'.text'));
         mainasmlist.concat(gennasmrec(DIRECT,S_NO,'.align 4'));
         asmlines:=0;
      end;

    function gennasmrec(i : tasmop;s : topsize;const t : string) : pasmrec;

      var
         p : pasmrec;

      begin
         if (i=PUSH) or (i=PUSHAD) then pushgened:=true;
         new(p);
         p^.linenr:=aktlinenr;
         p^.infile:=inputstack;
         p^.instruc:=i;
         p^.disposing:=D_STRING;
         p^.opsize:=s;
         p^.t:=strpnew(t);
         p^.next:=nil;
         gennasmrec:=p;
      end;

    function genpasmrec(i : tasmop;s : topsize;pc : pchar) : pasmrec;

      var
         p : pasmrec;

      begin
         if (i=PUSH) or (i=PUSHAD) then pushgened:=true;
         new(p);
         p^.linenr:=aktlinenr;
         p^.infile:=inputstack;
         p^.instruc:=i;
         p^.disposing:=D_STRING;
         p^.opsize:=s;
         p^.t:=strnew(pc);
         p^.next:=nil;
         genpasmrec:=p;
      end;

    function genlasmrec(i : tasmop;l : longint) : pasmrec;

      var
         p : pasmrec;

      begin
         new(p);
         p^.linenr:=aktlinenr;
         p^.infile:=inputstack;
         p^.instruc:=i;
         p^.opsize:=S_NO;
         p^.disposing:=D_LABEL;
         p^.l:=l;
         p^.next:=nil;
         genlasmrec:=p;
      end;

   { aus Geschwindigkeitsgrnden "static" }

   type
      tbuffer = array[1..65000] of char;

   var
      buffer : ^tbuffer;
      buffersize : longint;
      bufferpos : word;

   procedure writestring(const s : string);

     begin
        if bufferpos+length(s)+2>buffersize then
          begin
             blockwrite(asmfile,buffer^,bufferpos-1);
             bufferpos:=1;
          end;
        move(s[1],buffer^[bufferpos],length(s));
        inc(bufferpos,length(s));
     end;

   procedure writechararray(p : pchar);

     var
        w : word;

     begin
        w:=strlen(p);
        if bufferpos+w+2>buffersize then
          begin
             blockwrite(asmfile,buffer^,bufferpos-1);
             bufferpos:=1;
          end;
        move(p^,buffer^[bufferpos],w);
        inc(bufferpos,w);
     end;

   function tolabel(l : longint) : string;

     var
        s : string;

     begin
        str(l,s);
        tolabel:='L'+s;
     end;

   procedure writemainasmlist;

     var
        p : pasmrec;
        lastline : longint;
        s,hs : string;
        infile : pinputstack;

     begin
{$ifdef tp}
       {$I-}
       rewrite(asmfile,1);
       {$I+}
{$else}
       rewrite(asmfile,1);
{$endif}
       if ioresult<>0 then
         fatalerror(cannot_open_asmfile);
       lastline:=-1;
       infile:=nil;

       { Buffer allozieren }
       buffersize:=maxavail;
       if buffersize>65000 then
         buffersize:=65000;
       getmem(buffer,buffersize);
       bufferpos:=1;
       p:=mainasmlist.wurzel;
       while assigned(p) do
         begin
            { Debuggerinformationen }
            if (p^.linenr<>lastline) and (p^.linenr<>0) then
              begin
                 if writeasmfile then
                   begin
                      writestring(#13#10#9'// Zeile #'+tostr(p^.linenr));
                      inc(asmlines,2);
                   end;
                 if debuginfo then
                   begin
                      writestring(#13#10#9'.stabd'#9'68,0,'+tostr(p^.linenr));
                      inc(asmlines);
                   end;
                 lastline:=p^.linenr;
              end;
            if debuginfo and (p^.infile<>infile) then
              begin
                 if not(assigned(infile)) then
                   writestring(#13#10'.stabs "'+upper(p^.infile^.filename^)+'",100,0,0,Ltext0')
                 else
                   writestring(#13#10'.stabs "'+upper(p^.infile^.filename^)+'",132,0,0,Ltext0');
                 inc(asmlines);
                 infile:=p^.infile;
              end;
            { nun Anweisung schreiben }
            if p^.instruc=A_NONE then
              begin
                 p:=p^.next;
                 continue;
              end;
            if p^.instruc=A_LABEL then writestring(#13#10+tolabel(p^.l)+':')
            else if p^.disposing=D_LABEL then writestring(#13#10#9+op_2_asm[p^.instruc]+#9+tolabel(p^.l))
            else if p^.instruc=DIRECT then
              begin
                 writestring(#13#10);
                 writechararray(p^.t);
              end
            else if (p^.instruc=PUSH) and (p^.opsize=S_W) and (p^.t[0]='$') then
              begin
                 { Bug im GNU-Assembler umgehen: }
                 writestring(#13#10#9'.byte 0x66'#13#10#9'.byte 0x68');
                 writestring(#13#10#9'.word ');
                 writechararray(p^.t+1);
              end
            else if assigned(p^.t) then
               begin
                  writestring(#13#10#9+op_2_asm[p^.instruc]+postf_2_asm[p^.opsize]+#9);
                  writechararray(p^.t);
               end
            else writestring(#13#10#9+op_2_asm[p^.instruc]+postf_2_asm[p^.opsize]+#13#10);
            inc(asmlines);
            p:=p^.next;
         end;
       writestring(#13#10#26);
       blockwrite(asmfile,buffer^,bufferpos-1);
       freemem(buffer,buffersize);
       close(asmfile);
    end;

end.
